@@ -24,29 +24,26 @@
 
 #include "grd-context.h"
 
+#include "grd-credentials-file.h"
+#include "grd-credentials-libsecret.h"
+#include "grd-credentials-tpm.h"
 #include "grd-egl-thread.h"
+#include "grd-settings.h"
 
 #include "grd-dbus-remote-desktop.h"
 #include "grd-dbus-screen-cast.h"
 
-static const GDebugKey grd_debug_keys[] = {
-  { "vnc", GRD_DEBUG_VNC },
-};
-
 struct _GrdContext
 {
   GObject parent;
 
-  GMainContext *main_context;
-
   GrdDBusRemoteDesktop *remote_desktop_proxy;
   GrdDBusScreenCast *screen_cast_proxy;
 
   GrdEglThread *egl_thread;
 
+  GrdCredentials *credentials;
   GrdSettings *settings;
-
-  GrdDebugFlags debug_flags;
 };
 
 G_DEFINE_TYPE (GrdContext, grd_context, G_TYPE_OBJECT)
@@ -79,43 +76,69 @@ grd_context_set_screen_cast_proxy (GrdContext        *context,
   context->screen_cast_proxy = proxy;
 }
 
-GMainContext *
-grd_context_get_main_context (GrdContext *context)
-{
-  return context->main_context;
-}
-
 GrdSettings *
 grd_context_get_settings (GrdContext *context)
 {
   return context->settings;
 }
 
+GrdCredentials *
+grd_context_get_credentials (GrdContext *context)
+{
+  return context->credentials;
+}
+
 GrdEglThread *
 grd_context_get_egl_thread (GrdContext *context)
 {
   return context->egl_thread;
 }
 
-GrdDebugFlags
-grd_context_get_debug_flags (GrdContext *context)
+void
+grd_context_notify_daemon_ready (GrdContext *context)
 {
-  return context->debug_flags;
+  g_autoptr (GError) error = NULL;
+
+  if (context->egl_thread)
+    return;
+
+  context->egl_thread = grd_egl_thread_new (&error);
+  if (!context->egl_thread)
+    g_debug ("Failed to create EGL thread: %s", error->message);
 }
 
-static void
-init_debug_flags (GrdContext *context)
+GrdContext *
+grd_context_new (GrdRuntimeMode   runtime_mode,
+                 GError         **error)
 {
-  const char *debug_env;
+  g_autoptr (GrdContext) context = NULL;
+  g_autoptr (GError) local_error = NULL;
 
-  debug_env = g_getenv ("GNOME_REMOTE_DESKTOP_DEBUG");
-  if (debug_env)
+  context = g_object_new (GRD_TYPE_CONTEXT, NULL);
+
+  switch (runtime_mode)
     {
-      context->debug_flags =
-        g_parse_debug_string (debug_env,
-                              grd_debug_keys,
-                              G_N_ELEMENTS (grd_debug_keys));
+    case GRD_RUNTIME_MODE_HEADLESS:
+      context->credentials = GRD_CREDENTIALS (grd_credentials_tpm_new (&local_error));
+      if (!context->credentials)
+        {
+          g_warning ("Init TPM credentials failed because %s, using GKeyFile as fallback",
+                     local_error->message);
+          context->credentials =
+            GRD_CREDENTIALS (grd_credentials_file_new (error));
+        }
+      break;
+    case GRD_RUNTIME_MODE_SCREEN_SHARE:
+      context->credentials = GRD_CREDENTIALS (grd_credentials_libsecret_new ());
+      break;
     }
+
+  if (!context->credentials)
+    return NULL;
+
+  context->settings = grd_settings_new (context);
+
+  return g_steal_pointer (&context);
 }
 
 static void
@@ -127,6 +150,7 @@ grd_context_finalize (GObject *object)
   g_clear_object (&context->screen_cast_proxy);
   g_clear_pointer (&context->egl_thread, grd_egl_thread_free);
   g_clear_object (&context->settings);
+  g_clear_object (&context->credentials);
 
   G_OBJECT_CLASS (grd_context_parent_class)->finalize (object);
 }
@@ -134,17 +158,6 @@ grd_context_finalize (GObject *object)
 static void
 grd_context_init (GrdContext *context)
 {
-  g_autoptr (GError) error = NULL;
-
-  context->main_context = g_main_context_default ();
-
-  init_debug_flags (context);
-
-  context->settings = g_object_new (GRD_TYPE_SETTINGS, NULL);
-
-  context->egl_thread = grd_egl_thread_new (&error);
-  if (!context->egl_thread)
-    g_debug ("Failed to create EGL thread: %s", error->message);
 }
 
 static void
