@@ -31,6 +31,7 @@ struct _GrdRdpDisplayControl
   DispServerContext *disp_context;
   gboolean channel_opened;
   gboolean channel_unavailable;
+  gboolean pending_caps;
 
   uint32_t channel_id;
   uint32_t dvc_subscription_id;
@@ -80,6 +81,7 @@ dvc_creation_status (gpointer user_data,
 
   g_debug ("[RDP.DISP] Channel opened successfully. "
            "Ready for receiving new monitor layouts");
+  display_control->pending_caps = FALSE;
   disp_context->DisplayControlCaps (disp_context);
 }
 
@@ -109,25 +111,41 @@ disp_monitor_layout (DispServerContext                        *disp_context,
   GrdRdpDisplayControl *display_control = disp_context->custom;
   GrdSessionRdp *session_rdp = display_control->session_rdp;
   GrdRdpMonitorConfig *monitor_config;
+  g_autoptr (GError) error = NULL;
+
+  if (display_control->pending_caps)
+    {
+      g_warning ("[RDP.DISP] Protocol violation: Received monitor layout PDU "
+                 "before being able to send capabilities. Terminating session");
+      grd_session_rdp_notify_error (session_rdp,
+                                    GRD_SESSION_RDP_ERROR_BAD_MONITOR_DATA);
+      return CHANNEL_RC_OK;
+    }
 
   if (monitor_layout_pdu->NumMonitors > disp_context->MaxNumMonitors)
     {
-      g_warning ("[RDP.DISP] Received invalid monitor layout pdu "
-                 "(exceeded maximum monitor count). Ignoring PDU...");
+      g_warning ("[RDP.DISP] Protocol violation: Invalid monitor layout: "
+                 "(exceeded maximum monitor count). Terminating session");
+      grd_session_rdp_notify_error (session_rdp,
+                                    GRD_SESSION_RDP_ERROR_BAD_MONITOR_DATA);
+      return CHANNEL_RC_OK;
+    }
+
+  monitor_config =
+    grd_rdp_monitor_config_new_from_disp_monitor_layout (monitor_layout_pdu,
+                                                         &error);
+  if (!monitor_config)
+    {
+      g_warning ("[RDP.DISP] Received invalid monitor layout from client: %s. "
+                 "Terminating session", error->message);
+      grd_session_rdp_notify_error (session_rdp,
+                                    GRD_SESSION_RDP_ERROR_BAD_MONITOR_DATA);
       return CHANNEL_RC_OK;
     }
 
   g_debug ("[RDP.DISP] Received new monitor layout. PDU contains %i monitors",
            monitor_layout_pdu->NumMonitors);
 
-  monitor_config =
-    grd_rdp_monitor_config_new_from_disp_monitor_layout (monitor_layout_pdu);
-  if (!monitor_config)
-    {
-      g_warning ("[RDP.DISP] Monitor layout PDU is invalid. Ignoring PDU...");
-      return CHANNEL_RC_OK;
-    }
-
   grd_session_rdp_submit_new_monitor_config (session_rdp, monitor_config);
 
   return CHANNEL_RC_OK;
@@ -226,6 +244,8 @@ grd_rdp_display_control_init (GrdRdpDisplayControl *display_control)
 {
   GSource *channel_teardown_source;
 
+  display_control->pending_caps = TRUE;
+
   channel_teardown_source = g_source_new (&source_funcs, sizeof (GSource));
   g_source_set_callback (channel_teardown_source, tear_down_channel,
                          display_control, NULL);
