@@ -57,6 +57,12 @@ typedef void (* GrdRdpFrameReadyCallback) (GrdRdpPipeWireStream *stream,
                                            gboolean              success,
                                            gpointer              user_data);
 
+typedef struct
+{
+  GMutex buffer_mutex;
+  gboolean is_locked;
+} BufferContext;
+
 struct _GrdRdpFrame
 {
   gatomicrefcount refcount;
@@ -67,14 +73,6 @@ struct _GrdRdpFrame
   size_t map_size;
   uint8_t *map;
 
-  gboolean has_pointer_data;
-  uint8_t *pointer_bitmap;
-  uint16_t pointer_hotspot_x;
-  uint16_t pointer_hotspot_y;
-  uint16_t pointer_width;
-  uint16_t pointer_height;
-  gboolean pointer_is_hidden;
-
   GrdRdpPipeWireStream *stream;
   GrdRdpFrameReadyCallback callback;
   gpointer callback_user_data;
@@ -82,19 +80,25 @@ struct _GrdRdpFrame
 
 typedef struct
 {
-  GrdHwAccelNvidia *hwaccel_nvidia;
+  uint8_t *pointer_bitmap;
+  uint16_t pointer_hotspot_x;
+  uint16_t pointer_hotspot_y;
+  uint16_t pointer_width;
+  uint16_t pointer_height;
+} RdpPointer;
+
+typedef struct
+{
   GrdRdpBuffer *rdp_buffer;
 } UnmapBufferData;
 
 typedef struct
 {
-  GrdHwAccelNvidia *hwaccel_nvidia;
   GrdRdpBuffer *rdp_buffer;
 } AllocateBufferData;
 
 typedef struct
 {
-  GrdHwAccelNvidia *hwaccel_nvidia;
   GrdRdpBuffer *rdp_buffer;
 } RealizeBufferData;
 
@@ -112,6 +116,10 @@ struct _GrdRdpPipeWireStream
 
   GrdSessionRdp *session_rdp;
   GrdRdpSurface *rdp_surface;
+  GrdEglThreadSlot egl_slot;
+
+  GMutex dequeue_mutex;
+  gboolean dequeuing_disallowed;
 
   GSource *pipewire_source;
   struct pw_context *pipewire_context;
@@ -121,13 +129,19 @@ struct _GrdRdpPipeWireStream
 
   GrdRdpBufferPool *buffer_pool;
 
-  GSource *render_source;
+  GSource *frame_render_source;
   GMutex frame_mutex;
   GrdRdpFrame *pending_frame;
 
+  GSource *pointer_render_source;
+  GMutex pointer_mutex;
+  RdpPointer *pending_pointer;
+
   struct pw_stream *pipewire_stream;
   struct spa_hook pipewire_stream_listener;
 
+  GHashTable *pipewire_buffers;
+
   uint32_t src_node_id;
 
   struct spa_video_info_raw spa_format;
@@ -140,6 +154,57 @@ static void grd_rdp_frame_unref (GrdRdpFrame *frame);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdRdpFrame, grd_rdp_frame_unref)
 
+static BufferContext *
+buffer_context_new (void)
+{
+  BufferContext *buffer_context;
+
+  buffer_context = g_new0 (BufferContext, 1);
+  g_mutex_init (&buffer_context->buffer_mutex);
+
+  return buffer_context;
+}
+
+static void
+buffer_context_free (BufferContext *buffer_context)
+{
+  g_mutex_clear (&buffer_context->buffer_mutex);
+
+  g_free (buffer_context);
+}
+
+static void
+acquire_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
+                              struct pw_buffer     *buffer)
+{
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  g_mutex_lock (&buffer_context->buffer_mutex);
+  g_assert (!buffer_context->is_locked);
+  buffer_context->is_locked = TRUE;
+}
+
+static void
+maybe_release_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
+                                    struct pw_buffer     *buffer)
+{
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  if (!buffer_context->is_locked)
+    return;
+
+  buffer_context->is_locked = FALSE;
+  g_mutex_unlock (&buffer_context->buffer_mutex);
+}
+
 static GrdRdpFrame *
 grd_rdp_frame_new (GrdRdpPipeWireStream     *stream,
                    GrdRdpFrameReadyCallback  callback,
@@ -172,11 +237,17 @@ grd_rdp_frame_unref (GrdRdpFrame *frame)
       g_assert (!frame->has_map);
 
       g_clear_pointer (&frame->buffer, grd_rdp_buffer_release);
-      g_free (frame->pointer_bitmap);
       g_free (frame);
     }
 }
 
+static void
+rdp_pointer_free (RdpPointer *rdp_pointer)
+{
+  g_clear_pointer (&rdp_pointer->pointer_bitmap, g_free);
+  g_free (rdp_pointer);
+}
+
 void
 grd_rdp_pipewire_stream_resize (GrdRdpPipeWireStream *stream,
                                 GrdRdpVirtualMonitor *virtual_monitor)
@@ -202,7 +273,7 @@ grd_rdp_pipewire_stream_resize (GrdRdpPipeWireStream *stream,
 }
 
 static gboolean
-do_render (gpointer user_data)
+render_frame (gpointer user_data)
 {
   GrdRdpPipeWireStream *stream = GRD_RDP_PIPEWIRE_STREAM (user_data);
   GrdRdpFrame *frame;
@@ -226,21 +297,40 @@ do_render (gpointer user_data)
   grd_session_rdp_maybe_encode_new_frame (stream->session_rdp,
                                           stream->rdp_surface);
 
-  if (frame->pointer_bitmap)
+  grd_rdp_frame_unref (frame);
+
+  return G_SOURCE_CONTINUE;
+}
+
+static gboolean
+render_mouse_pointer (gpointer user_data)
+{
+  GrdRdpPipeWireStream *stream = user_data;
+  g_autoptr (GMutexLocker) locker = NULL;
+  RdpPointer *rdp_pointer;
+
+  locker = g_mutex_locker_new (&stream->pointer_mutex);
+  if (!stream->pending_pointer)
+    return G_SOURCE_CONTINUE;
+
+  rdp_pointer = g_steal_pointer (&stream->pending_pointer);
+  g_clear_pointer (&locker, g_mutex_locker_free);
+
+  if (rdp_pointer->pointer_bitmap)
     {
       grd_session_rdp_update_pointer (stream->session_rdp,
-                                      frame->pointer_hotspot_x,
-                                      frame->pointer_hotspot_y,
-                                      frame->pointer_width,
-                                      frame->pointer_height,
-                                      g_steal_pointer (&frame->pointer_bitmap));
+                                      rdp_pointer->pointer_hotspot_x,
+                                      rdp_pointer->pointer_hotspot_y,
+                                      rdp_pointer->pointer_width,
+                                      rdp_pointer->pointer_height,
+                                      g_steal_pointer (&rdp_pointer->pointer_bitmap));
     }
-  else if (frame->pointer_is_hidden)
+  else
     {
       grd_session_rdp_hide_pointer (stream->session_rdp);
     }
 
-  grd_rdp_frame_unref (frame);
+  rdp_pointer_free (rdp_pointer);
 
   return G_SOURCE_CONTINUE;
 }
@@ -261,13 +351,22 @@ static GSourceFuncs render_source_funcs =
 };
 
 static void
-create_render_source (GrdRdpPipeWireStream *stream,
-                      GMainContext         *render_context)
+create_render_sources (GrdRdpPipeWireStream *stream,
+                       GMainContext         *render_context)
 {
-  stream->render_source = g_source_new (&render_source_funcs, sizeof (GSource));
-  g_source_set_callback (stream->render_source, do_render, stream, NULL);
-  g_source_set_ready_time (stream->render_source, -1);
-  g_source_attach (stream->render_source, render_context);
+  stream->frame_render_source = g_source_new (&render_source_funcs,
+                                              sizeof (GSource));
+  g_source_set_callback (stream->frame_render_source,
+                         render_frame, stream, NULL);
+  g_source_set_ready_time (stream->frame_render_source, -1);
+  g_source_attach (stream->frame_render_source, render_context);
+
+  stream->pointer_render_source = g_source_new (&render_source_funcs,
+                                                sizeof (GSource));
+  g_source_set_callback (stream->pointer_render_source,
+                         render_mouse_pointer, stream, NULL);
+  g_source_set_ready_time (stream->pointer_render_source, -1);
+  g_source_attach (stream->pointer_render_source, render_context);
 }
 
 static void
@@ -406,13 +505,40 @@ on_stream_param_changed (void                 *user_data,
                            params, G_N_ELEMENTS (params));
 }
 
+static void
+on_stream_add_buffer (void             *user_data,
+                      struct pw_buffer *buffer)
+{
+  GrdRdpPipeWireStream *stream = user_data;
+
+  g_hash_table_insert (stream->pipewire_buffers, buffer, buffer_context_new ());
+}
+
+static void
+on_stream_remove_buffer (void             *user_data,
+                         struct pw_buffer *buffer)
+{
+  GrdRdpPipeWireStream *stream = user_data;
+  BufferContext *buffer_context = NULL;
+
+  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
+                                     NULL, (gpointer *) &buffer_context))
+    g_assert_not_reached ();
+
+  /* Ensure buffer is not locked any more */
+  g_mutex_lock (&buffer_context->buffer_mutex);
+  g_mutex_unlock (&buffer_context->buffer_mutex);
+
+  g_hash_table_remove (stream->pipewire_buffers, buffer);
+}
+
 static void
 process_mouse_pointer_bitmap (GrdRdpPipeWireStream *stream,
-                              struct spa_buffer    *buffer,
-                              GrdRdpFrame          *frame)
+                              struct spa_buffer    *buffer)
 {
   struct spa_meta_cursor *spa_meta_cursor;
   struct spa_meta_bitmap *spa_meta_bitmap;
+  RdpPointer *rdp_pointer = NULL;
   GrdPixelFormat format;
 
   spa_meta_cursor = spa_buffer_find_meta_data (buffer, SPA_META_Cursor,
@@ -435,20 +561,66 @@ process_mouse_pointer_bitmap (GrdRdpPipeWireStream *stream,
       uint8_t *buf;
 
       buf = SPA_MEMBER (spa_meta_bitmap, spa_meta_bitmap->offset, uint8_t);
-      frame->pointer_bitmap =
-        g_memdup2 (buf, spa_meta_bitmap->size.height *
-                        spa_meta_bitmap->stride);
-      frame->pointer_hotspot_x = spa_meta_cursor->hotspot.x;
-      frame->pointer_hotspot_y = spa_meta_cursor->hotspot.y;
-      frame->pointer_width = spa_meta_bitmap->size.width;
-      frame->pointer_height = spa_meta_bitmap->size.height;
-      frame->has_pointer_data = TRUE;
+
+      rdp_pointer = g_new0 (RdpPointer, 1);
+      rdp_pointer->pointer_bitmap =
+        g_memdup2 (buf, spa_meta_bitmap->size.height * spa_meta_bitmap->stride);
+      rdp_pointer->pointer_hotspot_x = spa_meta_cursor->hotspot.x;
+      rdp_pointer->pointer_hotspot_y = spa_meta_cursor->hotspot.y;
+      rdp_pointer->pointer_width = spa_meta_bitmap->size.width;
+      rdp_pointer->pointer_height = spa_meta_bitmap->size.height;
     }
   else if (spa_meta_bitmap)
     {
-      frame->pointer_is_hidden = TRUE;
-      frame->has_pointer_data = TRUE;
+      rdp_pointer = g_new0 (RdpPointer, 1);
     }
+
+  if (rdp_pointer)
+    {
+      g_mutex_lock (&stream->pointer_mutex);
+      g_clear_pointer (&stream->pending_pointer, rdp_pointer_free);
+
+      stream->pending_pointer = rdp_pointer;
+      g_mutex_unlock (&stream->pointer_mutex);
+
+      g_source_set_ready_time (stream->pointer_render_source, 0);
+    }
+}
+
+static void
+on_frame_ready (GrdRdpPipeWireStream *stream,
+                GrdRdpFrame          *frame,
+                gboolean              success,
+                gpointer              user_data)
+{
+  struct pw_buffer *buffer = user_data;
+
+  g_assert (frame);
+  g_assert (buffer);
+
+  if (frame->has_map)
+    {
+      munmap (frame->map, frame->map_size);
+      frame->has_map = FALSE;
+    }
+
+  if (!success)
+    goto out;
+
+  g_assert (frame->buffer);
+
+  g_mutex_lock (&stream->frame_mutex);
+  g_clear_pointer (&stream->pending_frame, grd_rdp_frame_unref);
+
+  stream->pending_frame = g_steal_pointer (&frame);
+  g_mutex_unlock (&stream->frame_mutex);
+
+  g_source_set_ready_time (stream->frame_render_source, 0);
+out:
+  pw_stream_queue_buffer (stream->pipewire_stream, buffer);
+  maybe_release_pipewire_buffer_lock (stream, buffer);
+
+  g_clear_pointer (&frame, grd_rdp_frame_unref);
 }
 
 static void
@@ -465,7 +637,7 @@ copy_frame_data (GrdRdpFrame *frame,
 
   for (y = 0; y < height; ++y)
     {
-      memcpy (((uint8_t *) buffer->local_data) + y * dst_stride,
+      memcpy (grd_rdp_buffer_get_local_data (buffer) + y * dst_stride,
               ((uint8_t *) src_data) + y * src_stride,
               width * 4);
     }
@@ -476,26 +648,18 @@ cuda_unmap_resource (gpointer user_data)
 {
   UnmapBufferData *data = user_data;
 
-  if (!data->rdp_buffer->mapped_cuda_pointer)
-    return TRUE;
-
-  grd_hwaccel_nvidia_unmap_cuda_resource (data->hwaccel_nvidia,
-                                          data->rdp_buffer->cuda_resource,
-                                          data->rdp_buffer->cuda_stream);
-  data->rdp_buffer->mapped_cuda_pointer = 0;
+  grd_rdp_buffer_unmap_cuda_resource (data->rdp_buffer);
 
   return TRUE;
 }
 
 static void
-unmap_cuda_resources (GrdEglThread     *egl_thread,
-                      GrdHwAccelNvidia *hwaccel_nvidia,
-                      GrdRdpBuffer     *rdp_buffer)
+unmap_cuda_resources (GrdEglThread *egl_thread,
+                      GrdRdpBuffer *rdp_buffer)
 {
   UnmapBufferData *data;
 
   data = g_new0 (UnmapBufferData, 1);
-  data->hwaccel_nvidia = hwaccel_nvidia;
   data->rdp_buffer = rdp_buffer;
 
   grd_egl_thread_run_custom_task (egl_thread,
@@ -509,15 +673,8 @@ cuda_allocate_buffer (gpointer user_data,
 {
   AllocateBufferData *data = user_data;
   GrdRdpBuffer *rdp_buffer = data->rdp_buffer;
-  gboolean success;
 
-  success = grd_hwaccel_nvidia_register_read_only_gl_buffer (data->hwaccel_nvidia,
-                                                             &rdp_buffer->cuda_resource,
-                                                             pbo);
-  if (success)
-    rdp_buffer->pbo = pbo;
-
-  return success;
+  return grd_rdp_buffer_register_read_only_gl_buffer (rdp_buffer, pbo);
 }
 
 static gboolean
@@ -534,13 +691,8 @@ cuda_map_resource (gpointer user_data)
 {
   RealizeBufferData *data = user_data;
   GrdRdpBuffer *rdp_buffer = data->rdp_buffer;
-  size_t mapped_size = 0;
 
-  return grd_hwaccel_nvidia_map_cuda_resource (data->hwaccel_nvidia,
-                                               rdp_buffer->cuda_resource,
-                                               &rdp_buffer->mapped_cuda_pointer,
-                                               &mapped_size,
-                                               rdp_buffer->cuda_stream);
+  return grd_rdp_buffer_map_cuda_resource (rdp_buffer);
 }
 
 static gboolean
@@ -565,11 +717,12 @@ on_framebuffer_ready (gboolean success,
 
 static void
 process_frame_data (GrdRdpPipeWireStream *stream,
-                    struct spa_buffer    *buffer,
-                    GrdRdpFrame          *frame)
+                    struct pw_buffer     *pw_buffer)
 {
-  GrdRdpFrameReadyCallback callback = frame->callback;
-  gpointer user_data = frame->callback_user_data;
+  struct spa_buffer *buffer = pw_buffer->buffer;
+  g_autoptr (GrdRdpFrame) frame = NULL;
+  GrdRdpFrameReadyCallback callback;
+  gpointer user_data;
   uint32_t drm_format;
   int bpp;
   int width;
@@ -587,6 +740,10 @@ process_frame_data (GrdRdpPipeWireStream *stream,
   grd_get_spa_format_details (stream->spa_format.format,
                               &drm_format, &bpp);
 
+  frame = grd_rdp_frame_new (stream, on_frame_ready, pw_buffer);
+  callback = frame->callback;
+  user_data = frame->callback_user_data;
+
   if (buffer->datas[0].type == SPA_DATA_MemFd)
     {
       GrdSession *session = GRD_SESSION (stream->session_rdp);
@@ -607,7 +764,7 @@ process_frame_data (GrdRdpPipeWireStream *stream,
       if (map == MAP_FAILED)
         {
           g_warning ("Failed to mmap buffer: %s", g_strerror (errno));
-          callback (stream, g_steal_pointer (&frame), TRUE, user_data);
+          callback (stream, g_steal_pointer (&frame), FALSE, user_data);
           return;
         }
       src_data = SPA_MEMBER (map, buffer->datas[0].mapoffset, uint8_t);
@@ -621,7 +778,7 @@ process_frame_data (GrdRdpPipeWireStream *stream,
           return;
         }
       rdp_buffer = frame->buffer;
-      pbo = rdp_buffer->pbo;
+      pbo = grd_rdp_buffer_get_pbo (rdp_buffer);
 
       if (stream->rdp_surface->needs_no_local_data &&
           src_stride == dst_stride)
@@ -643,7 +800,7 @@ process_frame_data (GrdRdpPipeWireStream *stream,
 
           munmap (map, size);
 
-          data_to_upload = rdp_buffer->local_data;
+          data_to_upload = grd_rdp_buffer_get_local_data (rdp_buffer);
         }
 
       if (!hwaccel_nvidia)
@@ -652,17 +809,17 @@ process_frame_data (GrdRdpPipeWireStream *stream,
           return;
         }
 
-      unmap_cuda_resources (egl_thread, hwaccel_nvidia, rdp_buffer);
+      unmap_cuda_resources (egl_thread, rdp_buffer);
 
       allocate_buffer_data = g_new0 (AllocateBufferData, 1);
-      allocate_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
       allocate_buffer_data->rdp_buffer = rdp_buffer;
 
       realize_buffer_data = g_new0 (RealizeBufferData, 1);
-      realize_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
       realize_buffer_data->rdp_buffer = rdp_buffer;
 
+      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_upload (egl_thread,
+                             stream->egl_slot,
                              pbo,
                              height,
                              dst_stride,
@@ -724,27 +881,26 @@ process_frame_data (GrdRdpPipeWireStream *stream,
         }
 
       if (!stream->rdp_surface->needs_no_local_data)
-        dst_data = frame->buffer->local_data;
+        dst_data = grd_rdp_buffer_get_local_data (frame->buffer);
 
       if (hwaccel_nvidia)
         {
-          unmap_cuda_resources (egl_thread, hwaccel_nvidia, rdp_buffer);
+          unmap_cuda_resources (egl_thread, rdp_buffer);
 
           iface.allocate = allocate_buffer;
           iface.realize = realize_buffer;
 
           import_buffer_data = g_new0 (ImportBufferData, 1);
-          import_buffer_data->allocate.hwaccel_nvidia = hwaccel_nvidia;
           import_buffer_data->allocate.rdp_buffer = rdp_buffer;
-
-          import_buffer_data->realize.hwaccel_nvidia = hwaccel_nvidia;
           import_buffer_data->realize.rdp_buffer = rdp_buffer;
 
           import_buffer_data->rdp_buffer = rdp_buffer;
         }
 
+      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_download (egl_thread,
-                               rdp_buffer->pbo,
+                               stream->egl_slot,
+                               grd_rdp_buffer_get_pbo (rdp_buffer),
                                height,
                                dst_stride,
                                hwaccel_nvidia ? &iface : NULL,
@@ -769,73 +925,19 @@ process_frame_data (GrdRdpPipeWireStream *stream,
     }
 }
 
-static void
-take_pointer_data_from (GrdRdpFrame *src_frame,
-                        GrdRdpFrame *dst_frame)
-{
-  g_assert (!dst_frame->pointer_bitmap);
-  dst_frame->pointer_bitmap = g_steal_pointer (&src_frame->pointer_bitmap);
-
-  dst_frame->pointer_hotspot_x = src_frame->pointer_hotspot_x;
-  dst_frame->pointer_hotspot_y = src_frame->pointer_hotspot_y;
-  dst_frame->pointer_width = src_frame->pointer_width;
-  dst_frame->pointer_height = src_frame->pointer_height;
-  dst_frame->pointer_is_hidden = src_frame->pointer_is_hidden;
-  dst_frame->has_pointer_data = TRUE;
-}
-
-static void
-on_frame_ready (GrdRdpPipeWireStream *stream,
-                GrdRdpFrame          *frame,
-                gboolean              success,
-                gpointer              user_data)
-{
-  struct pw_buffer *buffer = user_data;
-  GrdRdpFrame *pending_frame;
-
-  g_assert (frame);
-
-  if (frame->has_map)
-    {
-      munmap (frame->map, frame->map_size);
-      frame->has_map = FALSE;
-    }
-
-  if (!success)
-    goto out;
-
-  g_mutex_lock (&stream->frame_mutex);
-  pending_frame = g_steal_pointer (&stream->pending_frame);
-  if (pending_frame)
-    {
-      if (!frame->buffer && pending_frame->buffer)
-        frame->buffer = g_steal_pointer (&pending_frame->buffer);
-      if (!frame->has_pointer_data && pending_frame->has_pointer_data)
-        take_pointer_data_from (pending_frame, frame);
-
-      grd_rdp_frame_unref (pending_frame);
-    }
-  stream->pending_frame = g_steal_pointer (&frame);
-  g_mutex_unlock (&stream->frame_mutex);
-
-out:
-  if (buffer)
-    pw_stream_queue_buffer (stream->pipewire_stream, buffer);
-
-  g_source_set_ready_time (stream->render_source, 0);
-
-  g_clear_pointer (&frame, grd_rdp_frame_unref);
-}
-
 static void
 on_stream_process (void *user_data)
 {
   GrdRdpPipeWireStream *stream = GRD_RDP_PIPEWIRE_STREAM (user_data);
-  g_autoptr (GrdRdpFrame) frame = NULL;
+  g_autoptr (GMutexLocker) locker = NULL;
   struct pw_buffer *last_pointer_buffer = NULL;
   struct pw_buffer *last_frame_buffer = NULL;
   struct pw_buffer *next_buffer;
 
+  locker = g_mutex_locker_new (&stream->dequeue_mutex);
+  if (stream->dequeuing_disallowed)
+    return;
+
   while ((next_buffer = pw_stream_dequeue_buffer (stream->pipewire_stream)))
     {
       struct spa_meta_header *spa_meta_header;
@@ -876,31 +978,24 @@ on_stream_process (void *user_data)
   if (!last_pointer_buffer && !last_frame_buffer)
     return;
 
-  frame = grd_rdp_frame_new (stream, on_frame_ready, last_frame_buffer);
   if (last_pointer_buffer)
     {
-      process_mouse_pointer_bitmap (stream, last_pointer_buffer->buffer, frame);
+      process_mouse_pointer_bitmap (stream, last_pointer_buffer->buffer);
       if (last_pointer_buffer != last_frame_buffer)
         pw_stream_queue_buffer (stream->pipewire_stream, last_pointer_buffer);
     }
-
   if (!last_frame_buffer)
-    {
-      GrdRdpFrameReadyCallback callback = frame->callback;
-      gpointer callback_user_data = frame->callback_user_data;
+    return;
 
-      callback (stream, g_steal_pointer (&frame), TRUE, callback_user_data);
-      return;
-    }
-
-  process_frame_data (stream, last_frame_buffer->buffer,
-                      g_steal_pointer (&frame));
+  process_frame_data (stream, last_frame_buffer);
 }
 
 static const struct pw_stream_events stream_events = {
   PW_VERSION_STREAM_EVENTS,
   .state_changed = on_stream_state_changed,
   .param_changed = on_stream_param_changed,
+  .add_buffer = on_stream_add_buffer,
+  .remove_buffer = on_stream_remove_buffer,
   .process = on_stream_process,
 };
 
@@ -1100,9 +1195,12 @@ grd_rdp_pipewire_stream_new (GrdSessionRdp               *session_rdp,
   stream->rdp_surface = rdp_surface;
   stream->src_node_id = src_node_id;
 
+  if (egl_thread && !hwaccel_nvidia)
+    stream->egl_slot = grd_egl_thread_acquire_slot (egl_thread);
+
   pw_init (NULL, NULL);
 
-  create_render_source (stream, render_context);
+  create_render_sources (stream, render_context);
 
   pipewire_source = grd_attached_pipewire_source_new ("RDP", error);
   if (!pipewire_source)
@@ -1153,9 +1251,9 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
   GrdContext *context = grd_session_get_context (session);
   GrdEglThread *egl_thread;
 
-  /* Setting a PipeWire stream inactive will wait for the data thread to end */
-  if (stream->pipewire_stream)
-    pw_stream_set_active (stream->pipewire_stream, false);
+  g_mutex_lock (&stream->dequeue_mutex);
+  stream->dequeuing_disallowed = TRUE;
+  g_mutex_unlock (&stream->dequeue_mutex);
 
   egl_thread = grd_context_get_egl_thread (context);
   if (egl_thread && stream->pipewire_stream)
@@ -1171,29 +1269,48 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
       g_clear_pointer (&stream->pipewire_source, g_source_unref);
     }
 
-  if (stream->render_source)
+  if (stream->pointer_render_source)
     {
-      g_source_destroy (stream->render_source);
-      g_clear_pointer (&stream->render_source, g_source_unref);
+      g_source_destroy (stream->pointer_render_source);
+      g_clear_pointer (&stream->pointer_render_source, g_source_unref);
+    }
+  if (stream->frame_render_source)
+    {
+      g_source_destroy (stream->frame_render_source);
+      g_clear_pointer (&stream->frame_render_source, g_source_unref);
     }
 
   grd_rdp_damage_detector_invalidate_surface (stream->rdp_surface->detector);
+  g_clear_pointer (&stream->pending_pointer, rdp_pointer_free);
   g_clear_pointer (&stream->pending_frame, grd_rdp_frame_unref);
 
   release_all_buffers (stream);
   g_clear_object (&stream->buffer_pool);
 
+  g_mutex_clear (&stream->pointer_mutex);
   g_mutex_clear (&stream->frame_mutex);
+  g_mutex_clear (&stream->dequeue_mutex);
+
+  g_clear_pointer (&stream->pipewire_buffers, g_hash_table_unref);
 
   pw_deinit ();
 
+  if (egl_thread)
+    grd_egl_thread_release_slot (egl_thread, stream->egl_slot);
+
   G_OBJECT_CLASS (grd_rdp_pipewire_stream_parent_class)->finalize (object);
 }
 
 static void
 grd_rdp_pipewire_stream_init (GrdRdpPipeWireStream *stream)
 {
+  stream->pipewire_buffers =
+    g_hash_table_new_full (NULL, NULL,
+                           NULL, (GDestroyNotify) buffer_context_free);
+
+  g_mutex_init (&stream->dequeue_mutex);
   g_mutex_init (&stream->frame_mutex);
+  g_mutex_init (&stream->pointer_mutex);
 }
 
 static void
